﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.IO.Compression;
using System.Threading;
// The .NET zip capabilities are pretty bad up to v4.0, so a library makes more sense. .NET 4.5 has better zip capabilities, but there's no sense in using 4.5 if we don't need to as 4.0 is more common and we don't want people to have to upgrade just for that. 
using Ionic.Zip; 

namespace CryptoWalletBackupConsole
{
    class Program
    {
        static void Main(string[] args)
        {
            List<string> lines = new List<string>();
            string password = string.Empty;
            if (args.Length > 0)
            {
                if (File.Exists(args[0]))
                {
                    lines = CreateListFromFile(args[0]);
                }
                else
                {
                    // Use the default if the file doesn't exist. 
                    // This allows using a garbage file name like "asdf" as the first argument so that a password can be used more easily. 
                    // Incidentally, this is really "hacky" and kind of sucky, but it keeps the logic much simpler and avoids the need for a lot of parsing or using another library to parse the command line arguments. 
                    lines = CreateListFromFile("backup-file-list.txt");
                }

                // If we have a second argument, then it is the password.
                if (args.Length == 2)
                {
                    password = args[1];
                }
            }
            else
            {
                lines = CreateListFromFile("backup-file-list.txt");
            }
            // Do it!
            DoBackup(lines, password);
        }

        private static List<string> CreateListFromFile(string backupListFile)
        {
            List<string> lines = new List<string>();
            if (File.Exists(backupListFile))
            {
                string line;
                // Read the file and display it line by line.
                System.IO.StreamReader file = new System.IO.StreamReader(backupListFile);
                while ((line = file.ReadLine()) != null)
                {
                    lines.Add(line);
                }
                file.Close();
            }
            return lines;
        }

        private static bool DoBackup(List<string> backupList, string password)
        {
            string folderName = DateTime.Now.Year.ToString()
                + "-" + DateTime.Now.Month.ToString("00")
                + "-" + DateTime.Now.Day.ToString("00")
                + "_" + DateTime.Now.Hour.ToString("00")
                + "-" + DateTime.Now.Minute.ToString("00")
                + "-" + DateTime.Now.Second.ToString("00")
                + "-" + DateTime.Now.Millisecond.ToString("000");
            Console.WriteLine("Backups will be written to: " + folderName);

            Directory.CreateDirectory(folderName);

            // Copy all files into the folder.
            foreach (string item in backupList)
            {
                string fullPathName = item;
                string backupFilename = string.Empty;
                if (File.Exists(fullPathName))
                {
                    // Remove illegal characters.
                    string regexSearch = new string(Path.GetInvalidFileNameChars()) + new string(Path.GetInvalidPathChars());
                    Regex r = new Regex(string.Format("[{0}]", Regex.Escape(regexSearch)));
                    backupFilename = r.Replace(fullPathName, "-");

                    File.Copy(fullPathName, folderName + Path.DirectorySeparatorChar + backupFilename);
                    Console.WriteLine("Backed up: " + fullPathName);
                }
            }

            // Verify that all the files are there...
            foreach (string item in backupList)
            {
                string fullPathName = item;
                string backupFilename = string.Empty;
                if (File.Exists(fullPathName))
                {
                    // Remove illegal characters.
                    string regexSearch = new string(Path.GetInvalidFileNameChars()) + new string(Path.GetInvalidPathChars());
                    Regex r = new Regex(string.Format("[{0}]", Regex.Escape(regexSearch)));
                    backupFilename = r.Replace(fullPathName, "-");

                    if (!File.Exists(folderName + Path.DirectorySeparatorChar + backupFilename))
                    {
                        Console.WriteLine("FAILED TO VERIFY - " + folderName + Path.DirectorySeparatorChar + backupFilename);
                        try
                        {
                            // Try to copy it again. This should never happen, but... (crappy way to do it - a loop or other verification would be better)
                            File.Copy(fullPathName, folderName + Path.DirectorySeparatorChar + backupFilename);
                            // Give the copy operation a little time to happen. 
                            Thread.Sleep(50);
                        }
                        catch
                        {
                            // Realistically, this should never happen. 
                            Console.WriteLine("FAILED TO BACK UP " + fullPathName);
                            Console.WriteLine("        PLEASE MANUALLY VERIFY THE BACKUP.");
                        }
                    }
                    else
                    {
                        Console.WriteLine("VERIFIED - " + folderName + Path.DirectorySeparatorChar + backupFilename);
                    }
                }
            }

            // Zip the folder.
            using (ZipFile zip = new ZipFile())
            {
                // Password protect the file only if we have a password.
                if (!string.IsNullOrEmpty(password))
                {
                    zip.Password = password;
                }
                zip.AddDirectory(folderName);
                zip.Comment = "Backup created: " + System.DateTime.Now.ToString("G") + "\r\nCreated by Crypto Wallet Backup Console Version (http://cynic.me)";
                zip.Save(folderName + ".zip");
                // Pause a little to avoid the file not being created by the time we try to look for it. 
                Thread.Sleep(50);
                // Get rid of the folder if we have the ZIP file. 
                if (File.Exists(folderName + ".zip"))
                {
                    Directory.Delete(folderName, true);
                }

                // Remind the user that we password protected the file. 
                if (!string.IsNullOrEmpty(password))
                {
                    Console.WriteLine("******** IMPORTANT - ARCHIVE IS ENCRYPTED ********");
                    Console.WriteLine("Your backup is password protected. DO NOT LOSE YOUR PASSWORD.");
                }
            }

            return true;
        }
    }
}
